@@ -18,11 +18,11 @@ from cStringIO import StringIO
 import tokenize
 
 # TODO: support comment tokens inside xhpy_text (e.g. &#187;)
-# TODO: xhpy_text puts spaces in places they aren't needed
 
 # parser state
 token = None
 next = None
+line = None   # current line, for debugging
 
 # table mapping operators to symbol_base subclasses
 symbol_table = {}
@@ -42,13 +42,24 @@ ignore_whitespace = [False]
 # [x for x in L if (x if f(x) else y)] is not!
 disallow_if_else = [False]
 
+class XHPySyntaxError(SyntaxError):
+  def __init__(self, msg):
+    super(XHPySyntaxError, self).__init__("""\
+%s at line %d, column %d
+
+%s%s^
+""" % (msg, token.start[0], token.start[1], line, ' '*token.start[1]))
+    self.lineno = token.start[0]
+
 def single_quotify(s):
   return "'%s'" % s.replace('\'', '\\\'')
 
 def advance(id=None):
   global token
+  if token.type == tokenize.ENDMARKER:
+    raise XHPySyntaxError("advance() past end")
   if id and token.id != id:
-    raise SyntaxError("Expected %r" % id)
+    raise XHPySyntaxError("Expected '%s'" % id)
   token = next()
 
 def single_expression(rbp=0):
@@ -123,7 +134,7 @@ def xhpy_token_normalize():
     yield token.type, token.value
     advance()
   else:
-    raise SyntaxError('Expected valid XHPy attribute token')
+    raise XHPySyntaxError('Expected valid XHPy attribute token')
 
 def xhpy_attribute_name():
   attr_name = []
@@ -169,16 +180,11 @@ def xhpy_attribute():
     attr_type_token = tokenize.NUMBER, str(TYPE_ENUM)
     advance('(name)')
     attr_meta_tokens = list(xhpy_enum())
-  elif token.id == ':':
-    attr_type_token = tokenize.NUMBER, str(TYPE_OBJECT)
-    advance(':')
-    attr_meta_tokens = list(xhpy_tag_name())
   else:
-    raise SyntaxError('Expected XHPy attribute type')
+    attr_type_token = tokenize.NUMBER, str(TYPE_OBJECT)
+    attr_meta_tokens = list(single_expression())
 
-  attr_name = ''
-  if attr_type_token[1] != str(TYPE_OBJECT):
-    _x, attr_name = xhpy_attribute_name().next()
+  _x, attr_name = xhpy_attribute_name().next()
 
   if token.id == '=':
     advance('=')
@@ -186,7 +192,7 @@ def xhpy_attribute():
   elif token.id == '@':
     advance('@')
     if token.value != 'required':
-      raise SyntaxError('Expected XHPy attribute @required annotation')
+      raise XHPySyntaxError('Expected XHPy attribute @required annotation')
     attr_required_token = tokenize.NAME, 'True'
     advance('(name)')
 
@@ -285,7 +291,7 @@ def xhpy_child_atom():
       yield t
     advance(')')
   else:
-    raise SyntaxError('Expected valid XHPy child')
+    raise XHPySyntaxError('Expected valid XHPy child')
 
 def xhpy_child_term():
   yield tokenize.OP, '('
@@ -431,10 +437,10 @@ class symbol_base(object):
   value = None
 
   def nud(self):
-    raise SyntaxError("Syntax error (%r)." % self.id)
-  
+    raise XHPySyntaxError("Syntax error (%r)." % self.id)
+
   def led(self):
-    raise SyntaxError("Unknown operator (%r)." % self.id)
+    raise XHPySyntaxError("Unknown operator (%r)." % self.id)
 
   std = None
 
@@ -575,6 +581,12 @@ def nud(self):
   for t in single_expression():
     yield t
 
+# support function dict dereference with unary ** operation
+@method(symbol('**'))
+def nud(self):
+  for t in single_expression():
+    yield t
+
 # support xhpy class names with unary : operation
 @method(symbol(':'))
 def nud(self):
@@ -601,7 +613,7 @@ def nud(self, recursive=False):
   ignore_whitespace.append(True)
   tag_type, tag_open_name = xhpy_tag_name().next()
   yield tag_type, tag_open_name
-  yield tokenize.OP, '(' 
+  yield tokenize.OP, '('
   for t in xhpy_tag_attrs():
     yield t
   if token.id == '/':
@@ -622,7 +634,7 @@ def nud(self, recursive=False):
         advance('/')
         tag_type, tag_close_name = xhpy_tag_name().next()
         if tag_open_name != tag_close_name:
-          raise SyntaxError("Expected closing tag </%s>, got </%s>" % (tag_open_name, tag_close_name))
+          raise XHPySyntaxError("Expected closing tag </%s>, got </%s>" % (tag_open_name, tag_close_name))
         ignore_whitespace.pop()
         advance('>')
         break
@@ -643,9 +655,9 @@ def nud(self, recursive=False):
       yield tokenize.OP, ','
   yield tokenize.OP, ']'
   yield tokenize.OP, ')'
-  
+
 def xhpy_tag_name():
-  # drop 
+  # drop
   tag_name = list(symbol(':')().nud())[1:]
   yield tokenize.NAME, ''.join([t[1] for t in tag_name])
 
@@ -665,7 +677,7 @@ def xhpy_tag_attrs():
       yield token.type, token.value
       advance('(literal)')
     else:
-      raise SyntaxError('Expected XHPy attribute value')
+      raise XHPySyntaxError('Expected XHPy attribute value')
     yield tokenize.OP, ','
   yield tokenize.OP, '}'
 
@@ -684,9 +696,13 @@ def get_relpos(p2, p1):
 
 def xhpy_text():
   text = []
+  end = token.end
   for t in grab_until(lambda: token.id in ['{', '<']):
+    if text and token.start != end:
+      text.append(' ')
+    end = token.end
     text.append(t[1])
-  yield tokenize.STRING, "%r" % ' '.join(text).replace('\'', '\\\'')
+  yield tokenize.STRING, "%r" % ''.join(text).replace('\'', '\\\'')
 
 @method(symbol('('))
 def led(self):
@@ -768,7 +784,7 @@ def led(self):
           # L[:y:]
         # L[:y]
       # L[:]
-    else: 
+    else:
       for t in expression():
         yield t
       if token.id == ':':
@@ -874,7 +890,7 @@ def led(self):
 @method(symbol('.'))
 def led(self):
   if token.id != '(name)':
-    raise SyntaxError('Expected an attribute name')
+    raise XHPySyntaxError('Expected an attribute name')
   yield token.type, token.value
   advance()
 
@@ -896,7 +912,7 @@ def nud(self):
 @method(symbol('not'))
 def led(self):
   if token.id != 'in':
-    raise SyntaxError('Invalid syntax')
+    raise XHPySyntaxError('Invalid syntax')
   yield token.type, token.value
   advance()
   self.id = 'not in'
@@ -982,8 +998,9 @@ def std(self):
       break
     yield token.type, token.value
     advance('except')
-    for t in expression():
-      yield t
+    if token.id != ':':
+      for t in expression():
+        yield t
     yield token.type, token.value
     advance(':')
     for t in block():
@@ -1035,7 +1052,7 @@ def function_argument():
 def std(self):
   global ignore_whitespace
   if token.id != '(name)':
-    raise SyntaxError('Expected a function name')
+    raise XHPySyntaxError('Expected a function name')
   yield token.type, token.value
   advance('(name)')
   ignore_whitespace.append(True)
@@ -1087,28 +1104,10 @@ def std(self):
       yield t
   in_xhpy_class.pop()
 
-def class_name_fragment():
-  if token.id == ':':
-    advance(':')
-    tag_type, tag_name = xhpy_tag_name().next()
-    yield tokenize.NAME, tag_name
-  elif token.id == '(name)':
-    yield token.type, token.value
-    advance('(name)')
-
-def class_name():
-  for t in class_name_fragment():
-    yield t
-  while token.id == '.':
-    yield token.type, token.value
-    advance('.')
-    for t in class_name_fragment():
-      yield t
-
 @method(symbol('@'))
 def std(self):
   if token.id != '(name)':
-    raise SyntaxError('Expected a decorator call')
+    raise XHPySyntaxError('Expected a decorator call')
   for t in expression():
     yield t
   yield token.type, token.value
@@ -1135,7 +1134,7 @@ def std(self):
     for t in u.std():
       yield t
   else:
-    raise SyntaxError('Expected a decorated function or class')
+    raise XHPySyntaxError('Expected a decorated function or class')
 
 @method(symbol('print'))
 def std(self):
@@ -1278,7 +1277,7 @@ def import_target():
 
 def import_dotted_name():
   if token.id != '(name)':
-    raise SyntaxError('Expected import dotted name')
+    raise XHPySyntaxError('Expected import dotted name')
   yield token.type, token.value
   advance('(name)')
   while True:
@@ -1313,7 +1312,7 @@ def std(self):
 def argument_list():
   while True:
     if token.id != '(name)':
-      raise SyntaxError('Expected an argument name')
+      raise XHPySyntaxError('Expected an argument name')
     yield token.type, token.value
     advance()
     if token.id != ',':
@@ -1322,6 +1321,7 @@ def argument_list():
     advance(',')
 
 def tokenize_python(program):
+  global line
   type_map = {
     tokenize.NUMBER: '(literal)',
     tokenize.STRING: '(literal)',
@@ -1342,8 +1342,9 @@ def tokenize_python(program):
       yield type_map[t_type], t_value, t_start, t_end, t_type
       if t_type == tokenize.ENDMARKER:
         break
+      line = t_line
     except KeyError:
-      raise SyntaxError("Unexpected token %r" % (t,))
+      raise XHPySyntaxError("Unexpected token %r" % (t,))
 
 def tokenize_collapse_multiple_strings(program):
   string_token = None
@@ -1382,9 +1383,10 @@ def tokenize_collapse_multiple_newlines(program):
       if newline_token is None:
         symbol = symbol_table['(newline)']
         newline_token = symbol()
-        newline_token.value = t.value
+        newline_token.value = ''
         newline_token.start = t.start
         newline_token.type = t.type
+      newline_token.value += t.value
       newline_token.end = t.end
       continue
     elif newline_token is not None:
@@ -1412,7 +1414,7 @@ def tokenize_xhpy(program):
     elif id == '(operator)':
       symbol = symbol_table.get(value)
       if not symbol:
-        raise SyntaxError("Unknown operator (%r)" % id)
+        raise XHPySyntaxError("Unknown operator (%r)" % id)
       s = symbol()
     else:
       symbol = symbol_table[id]
